#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_hpdf.h"
#include <hpdf.h>

zend_class_entry *HPDF_ce;
zend_class_entry *HPDFPage_ce;

zend_function_entry HPDF_methods[] = {
	PHP_ME(HPDF, __construct, NULL, ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(HPDF, GetFont, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDF, AddPage, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDF, SetPassword, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDF, SaveToFile, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

zend_function_entry HPDFPage_methods[] = {
	PHP_ME(HPDFPage, __construct, NULL, ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(HPDFPage, SetSize, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, BeginText, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, SetFontAndSize, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, MoveTextPos, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, ShowText, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, EndText, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, GetWidth, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, GetHeight, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(HPDFPage, TextWidth, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

zend_module_entry hpdf_module_entry = {
	STANDARD_MODULE_HEADER,
	PHP_HPDF_EXTNAME,
	NULL,
	PHP_MINIT(hpdf),
	NULL,
	NULL,
	NULL,
	NULL,
	PHP_HPDF_VERSION,
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_HPDF
ZEND_GET_MODULE(hpdf)
#endif

int le_hpdf_font;

static void error_handler(HPDF_STATUS error, HPDF_STATUS detail, void *user_data)
{
	php_printf("ERROR: number=%04X, detail=%u\n", (HPDF_UINT)error, (HPDF_UINT)detail);
}

PHP_MINIT_FUNCTION(hpdf)
{
	zend_class_entry hpdf_class;
	zend_class_entry hpdfpage_class;

	INIT_CLASS_ENTRY(hpdf_class, "HPDF", HPDF_methods);
	HPDF_ce = zend_register_internal_class(&hpdf_class TSRMLS_CC);
	zend_declare_property_long(HPDF_ce, "__handle", sizeof("__handle"), -1, ZEND_ACC_PRIVATE TSRMLS_CC);

	INIT_CLASS_ENTRY(hpdfpage_class, "HPDFPage", HPDFPage_methods);
	HPDFPage_ce = zend_register_internal_class(&hpdfpage_class TSRMLS_CC);
	zend_declare_property_long(HPDFPage_ce, "__handle", sizeof("__handle"), -1, ZEND_ACC_PRIVATE TSRMLS_CC);
	zend_declare_class_constant_long(HPDFPage_ce, "SIZE_B5", sizeof("SIZE_B5")-1, HPDF_PAGE_SIZE_B5 TSRMLS_CC);
	zend_declare_class_constant_long(HPDFPage_ce, "LANDSCAPE", sizeof("LANDSCAPE")-1, HPDF_PAGE_LANDSCAPE TSRMLS_CC);

	le_hpdf_font = zend_register_list_destructors_ex(NULL, NULL, PHP_HPDF_FONT_RES_NAME, module_number);
}

static void *get_hpdf_handle(zval *hpdf)
{
	zval *zhpdf = zend_read_property(HPDF_ce, hpdf, "__handle", sizeof("__handle"), 1 TSRMLS_CC);
	return (void *)Z_LVAL_P(zhpdf);
}

static void *get_page_handle(zval *page)
{
	zval *zpage = zend_read_property(HPDFPage_ce, page, "__handle", sizeof("__handle"), 1 TSRMLS_CC);
	return (void *)Z_LVAL_P(zpage);
}

PHP_METHOD(HPDF, __construct)
{
	HPDF_Doc doc = HPDF_New(error_handler, NULL);
	if (!doc) {
		zend_throw_exception(zend_get_error_exception(), "HPDF_New failed", -1 TSRMLS_CC);
		return;
	}
	zend_update_property_long(EG(scope), getThis(), "__handle", sizeof("__handle"), (long)doc TSRMLS_CC);
}

PHP_METHOD(HPDFPage, __construct)
{
	zend_update_property_long(EG(scope), getThis(), "__handle", sizeof("__handle"), (long)-1 TSRMLS_CC);
}

PHP_METHOD(HPDF, GetFont)
{
	php_hpdf_font_t *font;
	char *fontname;
	int fontsize;
	HPDF_Doc doc;
	if (FAILURE == zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET, ZEND_NUM_ARGS() TSRMLS_CC, 
				"s", &fontname, &fontsize)){
		zend_throw_exception(zend_get_error_exception(), "Invalid font", -1 TSRMLS_CC);
		return;
	}
	doc = get_hpdf_handle(getThis());
	font = emalloc(sizeof(php_hpdf_font_t));
	font->handle = HPDF_GetFont(doc, fontname, NULL);
	if (!font->handle) {
		efree(font);
		RETURN_FALSE;
	}
	ZEND_REGISTER_RESOURCE(return_value, font, le_hpdf_font);
}

PHP_METHOD(HPDF, AddPage)
{
	HPDF_Page page;
	HPDF_Doc doc = get_hpdf_handle(getThis());
	page = HPDF_AddPage(doc);
	if (!page) {
		zend_throw_exception(zend_get_error_exception(), "HPDF_AddPage error", -1 TSRMLS_CC);
		return;
	}
	object_init_ex(return_value, HPDFPage_ce);
	return_value->refcount = 1;
	return_value->is_ref = 1;
	zend_update_property_long(HPDFPage_ce, return_value, "__handle", sizeof("__handle"), (long)page TSRMLS_CC);
}

PHP_METHOD(HPDFPage, SetSize)
{
	HPDF_Page page;
	long arg1;
	long arg2;
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ll", &arg1, &arg2)){
		zend_throw_exception(zend_get_error_exception(), "invalid argument", -1 TSRMLS_CC);
		return;
	}
	page = get_page_handle(getThis());
	HPDF_Page_SetSize(page, arg1, arg2);
	RETURN_TRUE;
}

PHP_METHOD(HPDFPage, BeginText)
{
	HPDF_Page page = get_page_handle(getThis());
	HPDF_Page_BeginText(page);
	RETURN_TRUE;
}

PHP_METHOD(HPDFPage, SetFontAndSize)
{
	HPDF_Page page;
	zval* zfont;
	php_hpdf_font_t *font;
	long size;
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", 
				&zfont, &size)){
		zend_throw_exception(zend_get_error_exception(), "invalid argument", -1 TSRMLS_CC);
		return;
	}
	page = get_page_handle(getThis());
	ZEND_FETCH_RESOURCE(font, php_hpdf_font_t*, &zfont, -1, PHP_HPDF_FONT_RES_NAME, le_hpdf_font);
	HPDF_Page_SetFontAndSize(page, font->handle, size);
	RETURN_TRUE;
}

PHP_METHOD(HPDFPage, TextWidth)
{
	HPDF_Page page;
	char *text;
	int textlen;
	float ret;
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", 
				&text, &textlen)){
		zend_throw_exception(zend_get_error_exception(), "invalid argument", -1 TSRMLS_CC);
		return;
	}
	page = get_page_handle(getThis());
	ret = HPDF_Page_TextWidth(page, text);
	RETURN_DOUBLE(ret);
}

PHP_METHOD(HPDFPage, GetWidth)
{
	HPDF_Page page;
	long ret;
	page = get_page_handle(getThis());
	ret = HPDF_Page_GetWidth(page);
	RETURN_LONG(ret);
}

PHP_METHOD(HPDFPage, GetHeight)
{
	HPDF_Page page;
	long ret;
	page = get_page_handle(getThis());
	ret = HPDF_Page_GetHeight(page);
	RETURN_LONG(ret);
}


PHP_METHOD(HPDFPage, MoveTextPos)
{
	HPDF_Page page;
	double x;
	double y;
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "dd", 
				&x, &y)){
		zend_throw_exception(zend_get_error_exception(), "invalid argument", -1 TSRMLS_CC);
		return;
	}
	page = get_page_handle(getThis());
	HPDF_Page_MoveTextPos(page, x, y);
	RETURN_TRUE;
}

PHP_METHOD(HPDFPage, ShowText)
{
	HPDF_Page page;
	char *text;
	int textlen;
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", 
				&text, &textlen)){
		zend_throw_exception(zend_get_error_exception(), "invalid argument", -1 TSRMLS_CC);
		return;
	}
	page = get_page_handle(getThis());
	HPDF_Page_ShowText(page, text);
	RETURN_TRUE;
}

PHP_METHOD(HPDFPage, EndText)
{
	HPDF_Page page = get_page_handle(getThis());
	HPDF_Page_EndText(page);
	RETURN_TRUE;
}

PHP_METHOD(HPDF, SetPassword)
{
	HPDF_Doc doc;
	char *pass1;
	int pass1len;
	char *pass2;
	int pass2len;
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, 
				"ss", &pass1, &pass1len,
				&pass2, &pass2len)){
		zend_throw_exception(zend_get_error_exception(), "invalid argument", -1 TSRMLS_CC);
		return;
	}
	doc = get_hpdf_handle(getThis());
	HPDF_SetPassword(doc, pass1, pass2);
	RETURN_TRUE;
}

PHP_METHOD(HPDF, SaveToFile)
{
	HPDF_Doc doc;
	char *file;
	int filelen;
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, 
				"s", &file, &filelen)){
		zend_throw_exception(zend_get_error_exception(), "invalid argument", -1 TSRMLS_CC);
		return;
	}
	doc = get_hpdf_handle(getThis());
	HPDF_SaveToFile(doc, file);
	RETURN_TRUE;
}

PHP_METHOD(HPDF, __destruct)
{
}

PHP_METHOD(HPDFPage, __destruct)
{
	HPDF_Free(get_hpdf_handle(getThis()));
}

